<?php
/**
 * @file
 * Drupal RestMini module REST client.
 */

// Include the SimpleComplex library.
include_once __DIR__ . '/vendor/simplecomplex/restmini/src/Client.php';

/**
 * Drupal 7 specialisation of composer library class SimpleComplex RestMini
 * Client.
 */
class RestMiniClient extends \SimpleComplex\RestMini\Client {

  /**
   * SSL CA Root Certificates bundle URL.
   *
   * Don't override this; functions use it (indirectly).
   *
   * @var string
   */
  const SSL_CACERTS_URL = 'http://curl.haxx.se/ca/cacert.pem';

  /**
   * Only applies to Drupal variables; environment variables use (parent)
   * lib_simplecomplex_restmini_client.
   *
   * @var string
   */
  const CONFIG_DOMAIN = 'restmini_client';

  /**
   * Default when no 'log_severity' option.
   *
   * @var string
   */
  const LOG_SEVERITY_DEFAULT = WATCHDOG_WARNING;

  /**
   * Default when no 'log_type' option.
   *
   * @var string
   */
  const LOG_TYPE_DEFAULT = 'restmini_client';

  /**
   * @var integer
   */
  protected static $errorCodeOffset = 1500;

  /**
   * Get config var.
   *
   * This implementation supports Drupal variables, as fallback to server
   * environment vars.
   *
   * @param string $domain
   *   Default: static::CONFIG_DOMAIN.
   * @param string $name
   * @param mixed $default
   *   Default: NULL.
   *
   * @return mixed
   *   String, unless no such var and arg default isn't string.
   */
  protected static function configGet($domain, $name, $default = NULL) {
    if (($val = parent::configGet($domain ? $domain : parent::CONFIG_DOMAIN, $name, FALSE)) !== FALSE) {
      return $val;
    }
    return variable_get(($domain ? $domain : static::CONFIG_DOMAIN) . '_' . $name, $default);
  }

  /**
   * @param string $str
   *
   * @return string
   */
  protected static function plaintext($str) {
    return check_plain($str);
  }

  /**
   * Get default certificates dir.
   *
   * @return string
   */
  protected static function certificateDir() {
    static $dir;
    if (!$dir) {
      $dir = variable_get('restmini_client_cacertsdir', 'private://certs/ca');
      if (strpos(' ' . $dir, 'private://')) {
        $dir = DRUPAL_ROOT . '/' .
          str_replace(
          // Support that the 'file_private_path' variable might have
          // trailing slash;
          // the core Media File system settings page allows with/without.
            array('private://', '//'),
            array(variable_get('file_private_path') . '/', '/'),
            $dir
          );
      }
    }

    return $dir;
  }

  /**
   * Uses optional PSR-3 logger and/or Inspect if applicable.
   *
   * @param integer $severity
   * @param string $message
   * @param \Exception|NULL $exception
   *   Ignored if no Inspect library.
   * @param mixed $variable
   *   Ignored if no Inspect library.
   *   Ignored if truthy arg $exception.
   */
  protected function log($severity, $message, $exception = NULL, $variable = NULL) {
    static $inspect = -1, $logger;
    // Check for Inspect, and whether this object was initialized with a PSR-3
    // logger (as option).
    if ($inspect == -1) {
      $inspect = module_exists('inspect');
      // Use PSR-3 logger; directly or as logger for Inspect.
      if (!empty($this->options['logger'])) {
        $logger = $this->options['logger'];
        if (!is_object($logger) || !method_exists($logger, 'log')) {
          $logger = NULL;
        }
      }
    }
    $logType = !empty($this->options['log_type']) ? $this->options['log_type'] : static::LOG_TYPE_DEFAULT;
    if ($inspect) {
      $opts = array(
        'type' => $logType,
        'message' => $message,
        'severity' => $severity,
        'wrappers' => 1,
      );
      if ($logger) {
        $opts['logger'] = $logger;
      }
      // Trace exception, or inspect variable.
      if ($exception) {
        Inspect::trace($exception, $opts);
      }
      else {
        Inspect::log($variable, $opts);
      }
    }
    else {
      if ($exception) {
        $message .= ': (' . $exception->getCode() . ') ' . $exception->getMessage()
          . ' @' . $exception->getFile() . ':' . $exception->getLine();
      }
      if ($logger) {
        $logger->log($severity, $message);
      }
      else {
        watchdog(
          $logType,
          (!$message && $exception) ? $exception->getMessage() : $message,
          array(),
          $severity
        );
      }
    }
  }

  /**
   * @param string $filename
   *   Default: empty (~ use server env. var or Drupal var).
   */
  public static function downloadCaCertBundle($filename = '') {
    $isCli = drupal_is_cli();
    $errors = array();

    $caBundleFile = $filename ? $filename : (variable_get('restmini_client_cacertsdir', 'private://certs/ca') . '/cacert.pem');
    if (strpos(' ' . $caBundleFile, 'private://')) {
      $caBundleFile = DRUPAL_ROOT . '/' .
        str_replace(
          // Support that the 'file_private_path' variable might have
          // trailing slash;
          // the core Media File system settings page allows with/without.
          array('private://', '//'),
          array(variable_get('file_private_path') . '/', '/'),
          $caBundleFile
        );
    }

    // Make directory if it doesn't exist.
    $fileExists = FALSE;
    $dir = dirname($caBundleFile);
    if (!file_exists($dir)) {
      if (!mkdir($dir, 0775, TRUE)) {
        $errors[] = $isCli ? dt('Failed to create directory.') : t('Failed to create the directory.');
      }
    }
    else {
      $fileExists = file_exists($caBundleFile);
    }

    // Get remote file.
    $source = variable_get('restmini_client_cacertssrc', self::SSL_CACERTS_URL);
    if (($caBundle = file_get_contents($source))) {
      if (strpos($caBundle, '-----BEGIN CERTIFICATE-----') && strpos($caBundle, '-----END CERTIFICATE-----')) {
        if (!$errors) {
          if (!file_put_contents($caBundleFile, $caBundle)) {
            $errors[] = $isCli ? dt('Failed to save the SSL CA certs bundle file to !file', array('!file' => $caBundleFile)) :
              t('Failed to save the SSL CA certs bundle to !file', array('!file' => $caBundleFile));
          }
        }
      }
      else {
        $replacers = array('!begin' => '-----BEGIN CERTIFICATE-----', '!end' => '-----END CERTIFICATE-----');
        $errors[] = $isCli ? dt('The downloaded file doesn\'t contain !begin and !end.', $replacers) :
          t('The downloaded SSL CA certs bundle file doesn\'t contain !begin and !end.', $replacers);
      }
    }
    else {
      $errors[] = $isCli ? dt('Failed to download the SSL CA certs file from !url.', array('!url' => $source)) :
        t('Failed to download the SSL CA certs bundle file from !url.', array('!url' => $source));
    }

    if (!$errors) {
      $message = $isCli ? dt('Downloaded new CA certs bundle file from !url.', array('!url' => $source)) :
        t('Downloaded new SSL CA certs bundle file from !url.', array('!url' => $source));
      watchdog(
        'restmini_client',
        $message,
        array(),
        WATCHDOG_INFO
      );
      if ($isCli) {
        return;
      }
      drupal_set_message($message, 'notice');
    }
    else {
      $errors = array_merge(
        $errors,
        array(
          $isCli ? dt('Failed to download new SSL CA certs bundle file from !url and save it locally.', array('!url' => $source)) :
            t('Failed to download new SSL CA certs bundle file from !url and save it locally.', array('!url' => $source)),
          $fileExists ? ($isCli ? dt('Will continue using the existing file.') : t('Will continue using the existing file.')) :
            ($isCli ? dt('And there\'s no existing file to use as fall-back.') : t('And there\'s no existing file to use as fall-back.'))
        )
      );
      watchdog(
        'restmini_client',
        join(' ', $errors),
        array(),
        $fileExists ? WATCHDOG_WARNING : WATCHDOG_ERROR
      );
      if ($isCli) {
        return;
      }
      drupal_set_message(join('<br/>', $errors), $fileExists ? 'warning' : 'error');
    }
  }

}
